import { ArgumentNameError } from "./exceptions";
import { buildArgExtraConfig, parseArgsArray } from "./utils/utils";
import { convertToTable, formatArgHelp, tabTableRows } from "./utils/help";
import { checkAllPositionalValuesUsed, checkEnoughPositionalValues, createValueValidator, validateArgConfig } from "./utils/validation";
import { createValueCaster } from "./utils/cast";
/**
 * A collection of parsed arguments.
 *
 * @category Classes
 */
export class ParsedArgumentsCollection {
    constructor() {
        /**
         * A collection of parsed positional arguments.
         */
        this.positionalArgs = {};
        /**
         * A collection of parsed optional arguments.
         */
        this.optionArgs = {};
    }
    /**
     * Returns all arguments as a record.
     */
    get positional() {
        return Object.assign({}, this.positionalArgs);
    }
    /**
     * Returns all arguments as a record.
     */
    get options() {
        return Object.assign({}, this.optionArgs);
    }
    /**
     * Adds a new argument to the collection.
     * @param name - The name of the argument.
     * @param value - The value of the argument.
     * @returns The instance of ParsedArgumentsCollection for chaining.
     */
    add(name, value) {
        const formattedName = this.formatName(name);
        if (this.isPositional(name)) {
            this.positionalArgs[formattedName] = value;
        }
        else {
            this.optionArgs[formattedName] = value;
        }
        return this;
    }
    /**
     * Retrieves an argument value.
     * @param name - The name of the argument.
     * @returns The value of the argument, or undefined if not found.
     */
    get(name) {
        var _a, _b;
        const formattedName = this.formatName(name);
        if (this.isPositional(name)) {
            return ((_a = this.positional[formattedName]) !== null && _a !== void 0 ? _a : undefined);
        }
        return ((_b = this.optionArgs[formattedName]) !== null && _b !== void 0 ? _b : undefined);
    }
    /**
     * Checks if an argument exists in the collection.
     * @param name - The name of the argument.
     * @returns True if the argument exists, otherwise false.
     */
    has(name) {
        var _a, _b;
        const formattedName = this.formatName(name);
        if (this.isPositional(name)) {
            return ((_a = this.positional[formattedName]) !== null && _a !== void 0 ? _a : undefined) !== undefined;
        }
        return ((_b = this.optionArgs[formattedName]) !== null && _b !== void 0 ? _b : undefined) !== undefined;
    }
    /**
     * Formats the argument name by removing leading dashes.
     * @param name - The argument name.
     * @returns The formatted name.
     */
    formatName(name) {
        return name.replace(/^--/, '');
    }
    /**
     * Checks if the argument is positional.
     * @param name - The argument name.
     * @returns True if the argument is positional, otherwise false.
     */
    isPositional(name) {
        return !name.startsWith('--');
    }
}
/**
 * Parser for command-line arguments.
 *
 * @category Classes
 *
 * @example
 * ```
 * const parser = new ArgsParser([
 *    {
 *        name: 'action',
 *        description: 'Action to perform',
 *        type: 'string',
 *    },
 *    {
 *        name: '--flag',
 *        description: 'Flag',
 *        alias: '-f',
 *        type: 'boolean',
 *        const: true,
 *        default: false,
 *    },
 *    {
 *        name: '--values',
 *        description: 'Values',
 *        alias: '-v',
 *        type: 'string',
 *        nargs: '*',
 *        default: [],
 *    }
 * ]);
 *
 * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
 * const parsedArgs = parser.parse(argv);
 *
 * console.log(parsedArgs.positional); // { action: 'make' }
 * console.log(parsedArgs.optional); // { flag: true, values: ['a', 'b', 'c'] }
 * console.log(parsedArgs.get('--values')); // ['a', 'b', 'c']
 * ```
 */
export class ArgsParser {
    /**
     * Constructs an instance of ArgsParser with the provided argument configurations.
     *
     * @param args - An array of argument configurations.
     */
    constructor(args = []) {
        /**
         * Maps argument names to their configuration.
         */
        this.argsMap = new Map();
        /**
         * Maps argument aliases to their corresponding names.
         */
        this.aliasMap = new Map();
        /**
         * Tracks used argument names and aliases to prevent duplicates.
         */
        this.usedArgs = new Set();
        for (const arg of args) {
            this.addArgument(arg);
        }
    }
    /**
     * Retrieves the help message.
     */
    get help() {
        const positionalArgs = this.getPositionalArguments();
        const optionArgs = this.getOptionArguments();
        const positionalArgsRows = [];
        for (const arg of positionalArgs) {
            positionalArgsRows.push(...formatArgHelp(arg));
        }
        const optionalArgsRows = [];
        for (const arg of optionArgs) {
            optionalArgsRows.push(...formatArgHelp(arg));
        }
        const result = [];
        if (positionalArgsRows.length > 0) {
            result.push(['Positional arguments:'], [''], ...tabTableRows(positionalArgsRows));
        }
        if (optionalArgsRows.length > 0) {
            result.push(['Options:'], [''], ...tabTableRows(optionalArgsRows));
        }
        return convertToTable(result, 1);
    }
    /**
     * Adds the help message for the arguments to the parser.
     *
     * @returns The updated parser.
     */
    addHelp() {
        return this.addArgument({
            name: '--help',
            alias: '-h',
            description: 'Show help',
            type: 'boolean',
            const: true,
            default: false,
        });
    }
    /**
     * Adds a new argument configuration to the parser.
     *
     * @param config - The argument configuration.
     * @returns The instance of ArgsParser for chaining.
     *
     * @example
     * ```
     * const parser = new ArgsParser();
     * parser.addArgument({
     *   name: '--flag',
     *   alias: '-f',
     *   type: 'boolean',
     *   const: true,
     *   default: false,
     * });
     * ```
     */
    addArgument(config) {
        validateArgConfig(config, this.usedArgs);
        this.argsMap.set(config.name, this.extendArgConfig(config));
        this.usedArgs.add(config.name);
        if (config.alias !== undefined) {
            this.aliasMap.set(config.alias, config.name);
            this.usedArgs.add(config.alias);
        }
        return this;
    }
    /**
     * Parses the given argument string and returns a collection of parsed arguments.
     *
     * @param argv - The argument string.
     * @returns A ParsedArgumentsCollection containing the parsed arguments.
     *
     * @example
     * ```
     * const parser = new ArgsParser([
     *    {
     *        name: 'action',
     *        description: 'Action to perform',
     *        type: 'string',
     *    },
     *    {
     *        name: '--flag',
     *        description: 'Flag',
     *        alias: '-f',
     *        type: 'boolean',
     *        const: true,
     *        default: false,
     *    },
     *    {
     *        name: '--values',
     *        description: 'Values',
     *        alias: '-v',
     *        type: 'string',
     *        nargs: '*',
     *        default: [],
     *    }
     * ]);
     *
     * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
     * const parsedArgs = parser.parse(argv);
     *
     * console.log(parsedArgs.positional); // { action: 'make' }
     * console.log(parsedArgs.optional); // { flag: true, values: ['a', 'b', 'c'] }
     * console.log(parsedArgs.get('--values')); // ['a', 'b', 'c']
     * ```
     */
    parse(argv) {
        const positionalArgs = this.getPositionalArguments();
        const optionalArgs = new Set(this.getOptionArguments());
        let [parsedPositional, parsedOptions] = parseArgsArray(argv);
        parsedPositional.reverse();
        const result = new ParsedArgumentsCollection();
        for (let i = 0; i < positionalArgs.length; ++i) {
            const argConfig = positionalArgs[i];
            checkEnoughPositionalValues(parsedPositional, argConfig);
            const toReadCount = this.getArgsCountToRead(argConfig, positionalArgs.slice(i + 1), parsedPositional);
            const value = [];
            for (let i = 0; i < toReadCount; ++i) {
                value.push(parsedPositional.pop());
            }
            result.add(argConfig.name, this.processArgValue(value, argConfig, value.length > 0));
        }
        checkAllPositionalValuesUsed(parsedPositional);
        for (const [key, value] of Object.entries(parsedOptions)) {
            const argConfig = this.getArgConfig(key);
            result.add(argConfig.name, this.processArgValue(value, argConfig, true));
            optionalArgs.delete(argConfig);
        }
        for (const argConfig of optionalArgs) {
            result.add(argConfig.name, this.processArgValue([], argConfig, false));
        }
        return result;
    }
    /**
     * Processes a single argument value.
     *
     * @param value - The value(s) being processed.
     * @param argConfig - The configuration of the argument.
     * @param isset - Whether the argument is set.
     *
     * @returns The processed value, or undefined if the argument is not set.
     */
    processArgValue(value, argConfig, isset) {
        const validator = createValueValidator(argConfig);
        const caster = createValueCaster(argConfig);
        validator.validateBeforeCast(value, isset);
        const result = caster.cast(value, isset);
        validator.validateAfterCast(result);
        return result;
    }
    /**
     * Retrieves the positional arguments.
     *
     * @returns An array of positional argument configurations.
     */
    getPositionalArguments() {
        return [...this.argsMap.values()].filter((x) => x.positional);
    }
    /**
     * Retrieves the optional arguments.
     *
     * @returns An array of optional argument configurations.
     */
    getOptionArguments() {
        return [...this.argsMap.values()].filter((x) => !x.positional);
    }
    /**
     * Retrieves the argument configuration for a given key.
     *
     * @param key - The argument key.
     *
     * @returns The corresponding ArgConfigExtended.
     *
     * @throws ArgumentNameError if the argument key is unknown.
     */
    getArgConfig(key) {
        var _a;
        const config = this.argsMap.get((_a = this.aliasMap.get(key)) !== null && _a !== void 0 ? _a : key);
        if (config === undefined) {
            throw new ArgumentNameError(`Unknown argument ${key}.`);
        }
        return config;
    }
    /**
     * Extends an argument configuration with extra information.
     *
     * @param config - The argument configuration.
     *
     * @returns The argument configuration extended with extra information.
     */
    extendArgConfig(config) {
        return Object.assign(Object.assign({}, config), buildArgExtraConfig(config));
    }
    /**
     * Determines the number of arguments to read based on the nargs configuration.
     *
     * @param argConfig - The configuration specifying how many arguments can be read.
     * @param remainingArgConfigs - The remaining arguments to read.
     * @param values - The total number of available arguments.
     *
     * @returns The number of arguments to read.
     */
    getArgsCountToRead(argConfig, remainingArgConfigs, values) {
        const minRemainingValuesCount = remainingArgConfigs.reduce((acc, x) => acc + x.minValuesCount, 0);
        let maxCurrentValuesCount = values.length - minRemainingValuesCount;
        if (maxCurrentValuesCount < 0) {
            maxCurrentValuesCount = values.length;
        }
        if (!argConfig.multiple) {
            return Math.min(1, maxCurrentValuesCount);
        }
        if (argConfig.valuesCount !== undefined) {
            return Math.min(argConfig.valuesCount, maxCurrentValuesCount);
        }
        return maxCurrentValuesCount;
    }
}
//# sourceMappingURL=classes.js.map