import type { ArgConfig, ArgsParserInterface, ParsedArgumentsCollectionInterface } from "./types";
/**
 * A collection of parsed arguments.
 *
 * @category Classes
 */
export declare class ParsedArgumentsCollection implements ParsedArgumentsCollectionInterface {
    /**
     * A collection of parsed positional arguments.
     */
    private readonly positionalArgs;
    /**
     * A collection of parsed optional arguments.
     */
    private readonly optionArgs;
    /**
     * Returns all arguments as a record.
     */
    get positional(): Record<string, unknown>;
    /**
     * Returns all arguments as a record.
     */
    get options(): Record<string, unknown>;
    /**
     * Adds a new argument to the collection.
     * @param name - The name of the argument.
     * @param value - The value of the argument.
     * @returns The instance of ParsedArgumentsCollection for chaining.
     */
    add(name: string, value: unknown): this;
    /**
     * Retrieves an argument value.
     * @param name - The name of the argument.
     * @returns The value of the argument, or undefined if not found.
     */
    get<T = unknown>(name: string): T;
    /**
     * Checks if an argument exists in the collection.
     * @param name - The name of the argument.
     * @returns True if the argument exists, otherwise false.
     */
    has(name: string): boolean;
    /**
     * Formats the argument name by removing leading dashes.
     * @param name - The argument name.
     * @returns The formatted name.
     */
    private formatName;
    /**
     * Checks if the argument is positional.
     * @param name - The argument name.
     * @returns True if the argument is positional, otherwise false.
     */
    private isPositional;
}
/**
 * Parser for command-line arguments.
 *
 * @category Classes
 *
 * @example
 * ```
 * const parser = new ArgsParser([
 *    {
 *        name: 'action',
 *        description: 'Action to perform',
 *        type: 'string',
 *    },
 *    {
 *        name: '--flag',
 *        description: 'Flag',
 *        alias: '-f',
 *        type: 'boolean',
 *        const: true,
 *        default: false,
 *    },
 *    {
 *        name: '--values',
 *        description: 'Values',
 *        alias: '-v',
 *        type: 'string',
 *        nargs: '*',
 *        default: [],
 *    }
 * ]);
 *
 * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
 * const parsedArgs = parser.parse(argv);
 *
 * console.log(parsedArgs.positional); // { action: 'make' }
 * console.log(parsedArgs.optional); // { flag: true, values: ['a', 'b', 'c'] }
 * console.log(parsedArgs.get('--values')); // ['a', 'b', 'c']
 * ```
 */
export declare class ArgsParser implements ArgsParserInterface {
    /**
     * Maps argument names to their configuration.
     */
    private readonly argsMap;
    /**
     * Maps argument aliases to their corresponding names.
     */
    private readonly aliasMap;
    /**
     * Tracks used argument names and aliases to prevent duplicates.
     */
    private readonly usedArgs;
    /**
     * Constructs an instance of ArgsParser with the provided argument configurations.
     *
     * @param args - An array of argument configurations.
     */
    constructor(args?: ArgConfig[]);
    /**
     * Retrieves the help message.
     */
    get help(): string;
    /**
     * Adds the help message for the arguments to the parser.
     *
     * @returns The updated parser.
     */
    addHelp(): ArgsParser;
    /**
     * Adds a new argument configuration to the parser.
     *
     * @param config - The argument configuration.
     * @returns The instance of ArgsParser for chaining.
     *
     * @example
     * ```
     * const parser = new ArgsParser();
     * parser.addArgument({
     *   name: '--flag',
     *   alias: '-f',
     *   type: 'boolean',
     *   const: true,
     *   default: false,
     * });
     * ```
     */
    addArgument(config: ArgConfig): ArgsParser;
    /**
     * Parses the given argument string and returns a collection of parsed arguments.
     *
     * @param argv - The argument string.
     * @returns A ParsedArgumentsCollection containing the parsed arguments.
     *
     * @example
     * ```
     * const parser = new ArgsParser([
     *    {
     *        name: 'action',
     *        description: 'Action to perform',
     *        type: 'string',
     *    },
     *    {
     *        name: '--flag',
     *        description: 'Flag',
     *        alias: '-f',
     *        type: 'boolean',
     *        const: true,
     *        default: false,
     *    },
     *    {
     *        name: '--values',
     *        description: 'Values',
     *        alias: '-v',
     *        type: 'string',
     *        nargs: '*',
     *        default: [],
     *    }
     * ]);
     *
     * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
     * const parsedArgs = parser.parse(argv);
     *
     * console.log(parsedArgs.positional); // { action: 'make' }
     * console.log(parsedArgs.optional); // { flag: true, values: ['a', 'b', 'c'] }
     * console.log(parsedArgs.get('--values')); // ['a', 'b', 'c']
     * ```
     */
    parse(argv: string[]): ParsedArgumentsCollection;
    /**
     * Processes a single argument value.
     *
     * @param value - The value(s) being processed.
     * @param argConfig - The configuration of the argument.
     * @param isset - Whether the argument is set.
     *
     * @returns The processed value, or undefined if the argument is not set.
     */
    private processArgValue;
    /**
     * Retrieves the positional arguments.
     *
     * @returns An array of positional argument configurations.
     */
    private getPositionalArguments;
    /**
     * Retrieves the optional arguments.
     *
     * @returns An array of optional argument configurations.
     */
    private getOptionArguments;
    /**
     * Retrieves the argument configuration for a given key.
     *
     * @param key - The argument key.
     *
     * @returns The corresponding ArgConfigExtended.
     *
     * @throws ArgumentNameError if the argument key is unknown.
     */
    private getArgConfig;
    /**
     * Extends an argument configuration with extra information.
     *
     * @param config - The argument configuration.
     *
     * @returns The argument configuration extended with extra information.
     */
    private extendArgConfig;
    /**
     * Determines the number of arguments to read based on the nargs configuration.
     *
     * @param argConfig - The configuration specifying how many arguments can be read.
     * @param remainingArgConfigs - The remaining arguments to read.
     * @param values - The total number of available arguments.
     *
     * @returns The number of arguments to read.
     */
    private getArgsCountToRead;
}
