"use strict";
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseArgsArray = parseArgsArray;
exports.formatArgNameWithAlias = formatArgNameWithAlias;
exports.buildArgExtraConfig = buildArgExtraConfig;
/**
 * Parses an array of command-line arguments into positional and optional arguments.
 *
 * @param argv - The array of command-line arguments.
 * @returns A tuple containing an array of positional arguments and a record of optional arguments.
 *
 * @category Utils
 *
 * @example
 * ```
 * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
 * const [positional, optional] = parseArgsArray(argv);
 * console.log(positional); // ['make']
 * console.log(optional); // { '--flag': [], '-v': ['a', 'b', 'c'] }
 * ```
 */
function parseArgsArray(argv) {
    var foundIndex = argv.findIndex(function (x) { return x.startsWith('-'); });
    var optionalBegin = foundIndex !== -1 ? foundIndex : argv.length;
    var positional = __spreadArray([], __read(argv.slice(0, optionalBegin)), false);
    var optionalBuffer = __spreadArray([], __read(argv.slice(optionalBegin)), false).reverse();
    var optional = {};
    var argName = undefined;
    var argValues = [];
    while (optionalBuffer.length > 0) {
        var item = optionalBuffer.pop();
        if (item.match(/^-[a-zA-Z0-9]{2,}$/)) {
            var items = item.slice(1).split('').reverse();
            optionalBuffer.push.apply(optionalBuffer, __spreadArray([], __read(items.map(function (x) { return "-".concat(x); })), false));
            continue;
        }
        if (item.startsWith('-')) {
            if (argName !== undefined) {
                optional[argName] = argValues;
            }
            argName = item;
            argValues = [];
        }
        else {
            argValues.push(item);
        }
    }
    if (argName !== undefined) {
        optional[argName] = argValues;
    }
    return [positional, optional];
}
/**
 * Formats the argument name with alias.
 *
 * @param argConfig - The argument configuration.
 *
 * @returns The formatted argument name.
 *
 * @category Utils
 */
function formatArgNameWithAlias(argConfig) {
    return "".concat(argConfig.name).concat(argConfig.alias ? " (".concat(argConfig.alias, ")") : '');
}
/**
 * Builds the extra configuration for an argument.
 *
 * @param config - The argument configuration.
 *
 * @returns The extra configuration.
 *
 * @category Utils
 */
function buildArgExtraConfig(config) {
    var _a;
    var positional = !config.name.startsWith('--');
    var multiple = config.nargs === '*' || config.nargs === '+' || typeof config.nargs === 'number';
    var required = (_a = config.required) !== null && _a !== void 0 ? _a : (config.nargs !== '*' && config.nargs !== '?' && config.default === undefined);
    var allowEmpty = config.nargs === '*' || config.nargs === '?' || config.const !== undefined || (positional && !required);
    var valuesCount = typeof config.nargs === 'number' ? config.nargs : undefined;
    var minValuesCount = valuesCount !== null && valuesCount !== void 0 ? valuesCount : (allowEmpty ? 0 : 1);
    return { positional: positional, multiple: multiple, required: required, allowEmpty: allowEmpty, valuesCount: valuesCount, minValuesCount: minValuesCount };
}
//# sourceMappingURL=utils.js.map